import os
from typing import Iterator

from .common import Browser

# import winreg

# tuple of possible browsers
POSSIBLE_BROWSERS = {
    "Google Chrome": "chrome",
    "Google Chrome Canary": "chrome-canary",
    "Chromium": "chromium",
    "Mozilla Firefox": "firefox",
    "Firefox Developer Edition": "firefox-developer",
    "Firefox Nightly": "firefox-nightly",
    "Opera Stable": "opera",
    "Opera beta": "opera-beta",
    "Opera developer": "opera-developer",
    "Microsoft Edge": "msedge",
    "Microsoft Edge Beta": "msedge-beta",
    "Microsoft Edge Dev": "msedge-dev",
    "Microsoft Edge Canary": "msedge-canary",
    "Internet Explorer": "msie",
    "Brave": "brave",
    "Brave Beta": "brave-beta",
    "Brave Nightly": "brave-nightly",
}


# get all installed browsers
def browsers() -> Iterator[Browser]:
    import winreg
    yield from _win32_browsers_from_registry(winreg.HKEY_CURRENT_USER, winreg.KEY_READ)
    yield from _win32_browsers_from_registry(winreg.HKEY_LOCAL_MACHINE,
                                             winreg.KEY_READ | winreg.KEY_WOW64_64KEY)
    yield from _win32_browsers_from_registry(winreg.HKEY_LOCAL_MACHINE,
                                             winreg.KEY_READ | winreg.KEY_WOW64_32KEY)


# get browsers from registry
def _win32_browsers_from_registry(tree: int, access: int) -> Iterator[Browser]:
    import winreg
    try:
        with winreg.OpenKey(tree, r"Software\Clients\StartMenuInternet", access=access) as hkey:
            i = 0
            while True:
                try:
                    subkey = winreg.EnumKey(hkey, i)
                    i += 1
                except OSError:
                    break
                try:
                    description = winreg.QueryValue(hkey, subkey)
                    if not description or not isinstance(description, str):  # pragma: no cover
                        description = subkey
                except OSError:  # pragma: no cover
                    description = subkey
                try:
                    cmd = winreg.QueryValue(hkey, rf"{subkey}\shell\open\command")
                    cmd = cmd.strip('"')
                    os.stat(cmd)
                except (OSError, AttributeError, TypeError, ValueError):  # pragma: no cover
                    continue
                yield Browser(
                    name=POSSIBLE_BROWSERS.get(description, "unknown"),
                    description=description,
                    version="Not yet supported.",
                    # version=_get_file_version(cmd),
                    location=cmd
                )
    except FileNotFoundError:
        pass


# get browser version
# def _get_file_version(path: str) -> str:
#     import win32api
#
#     info = win32api.GetFileVersionInfo(path, "\\")
#     ms = info["FileVersionMS"]
#     ls = info["FileVersionLS"]
#     return ".".join(map(str, (win32api.HIWORD(ms), win32api.LOWORD(ms), win32api.HIWORD(ls), win32api.LOWORD(ls))))



import sys

if sys.version_info >= (3, 8):
    from typing import TypedDict
else:
    from typing_extensions import TypedDict


class OS:
    LINUX = "linux"
    WINDOWS = "win32"
    MAC = "darwin"


class Browser(TypedDict):
    name: str
    description: str
    version: str
    location: str


class Version(TypedDict):
    version: str



import os
import re
import subprocess
from typing import Iterator, Optional

from xdg.DesktopEntry import DesktopEntry

from .common import Browser, Version

# tuple of possible browsers
# desktop entry name may be different: "chromium_chromium.desktop" or "chromium-browser.desktop"
POSSIBLE_BROWSERS = (
    ("chrome", ("google-chrome",)),
    ("chromium", ("chromium", "chromium_chromium", "chromium-browser")),
    ("firefox", ("firefox", "firefox_firefox")),
    ("ms-edge", ("microsoft-edge",)),
    ("opera", ("opera_opera",)),
    ("opera-beta", ("opera-beta_opera-beta",)),
    ("opera-developer", ("opera-developer_opera-developer",)),
    ("brave", ("brave-browser", "brave_brave")),
    ("brave-beta", ("brave-browser-beta",)),
    ("brave-nightly", ("brave-browser-nightly",))
)

# tuple of browser locations
# $XDG_DATA_HOME and $XDG_DATA_DIRS are not always set
BROWSER_LOCATIONS = (
    "~/.local/share/applications",
    "/usr/share/applications",
    "/var/lib/snapd/desktop/applications",
)

# set version pattern with dot separation
VERSION_PATTERN = re.compile(r"\b(\S+\.\S+)\b")


# get all installed browsers
def browsers() -> Iterator[Browser]:
    for browser, desktop_entries in POSSIBLE_BROWSERS:
        for application_dir in BROWSER_LOCATIONS:
            for desktop_entry in desktop_entries:
                path = os.path.join(application_dir, f"{desktop_entry}.desktop")
                if not os.path.isfile(path):
                    continue
                entry = DesktopEntry(path)
                executable_path = entry.getExec()
                if executable_path.lower().endswith(" %u"):
                    executable_path = executable_path[:-3].strip()
                version = subprocess.getoutput(f"{executable_path} --version 2>&1").strip()
                match = VERSION_PATTERN.search(version)
                if match:
                    version = match[0]
                yield Browser(
                    name=browser, description=entry.getName(), version=version, location=executable_path
                )


# get details of a browser
def get_details_of(name) -> Optional[Browser]:
    for browser, desktop_names in (browser_record for browser_record in POSSIBLE_BROWSERS
                                   if browser_record[0] == name):
        for desktop_name in desktop_names:
            for application_dir in BROWSER_LOCATIONS:
                path = os.path.join(application_dir, f"{desktop_name}.desktop")
                if not os.path.isfile(path):
                    continue
                entry = DesktopEntry(path)
                executable_path = entry.getExec()
                if executable_path.lower().endswith(" %u"):
                    executable_path = executable_path[:-3].strip()
                version = subprocess.getoutput(f"{executable_path} --version 2>&1").strip()
                match = VERSION_PATTERN.search(version)
                if match:
                    version = match[0]
                yield Browser(
                    name=browser, description=entry.getName(), version=version, location=executable_path
                )


# retrieve browser version
def get_version_of(name) -> Optional[Version]:
    for browser, desktop_names in (browser_record for browser_record in POSSIBLE_BROWSERS
                                   if browser_record[0] == name):
        for desktop_name in desktop_names:
            for application_dir in BROWSER_LOCATIONS:
                path = os.path.join(application_dir, f"{desktop_name}.desktop")
                if not os.path.isfile(path):
                    continue
                entry = DesktopEntry(path)
                executable_path = entry.getExec()
                if executable_path.lower().endswith(" %u"):
                    executable_path = executable_path[:-3].strip()
                version = subprocess.getoutput(f"{executable_path} --version 2>&1").strip()
                match = VERSION_PATTERN.search(version)
                if match:
                    version = match[0]
                yield Version(
                    version=version
                )
    yield "Browser is not installed."


# check if the given browser is installed
def do_i_have_installed(name):
    for browser, desktop_names in (browser_record for browser_record in POSSIBLE_BROWSERS
                                   if browser_record[0] == name):
        for desktop_name in desktop_names:
            for application_dir in BROWSER_LOCATIONS:
                path = os.path.join(application_dir, f"{desktop_name}.desktop")
                if not os.path.isfile(path):
                    continue
                entry = DesktopEntry(path)
                if entry:
                    return True
    return False


# get default browser
def what_is_the_default_browser() -> Optional[str]:
    cmd = "xdg-settings get default-web-browser".split()
    if cmd:
        default_browser = subprocess.check_output(cmd, stderr=subprocess.DEVNULL).decode().strip()
    else:
        default_browser = "No browser is set to default."
    return default_browser



import os
import plistlib
import subprocess
from typing import Iterator, Optional

from .common import Browser, Version

# tuple of possible browsers
POSSIBLE_BROWSERS = (
    ("chrome", "com.google.Chrome", "KSVersion"),
    ("chrome-canary", "com.google.Chrome.canary", "KSVersion"),
    ("chromium", "org.chromium.Chromium", "CFBundleShortVersionString"),
    ("firefox", "org.mozilla.firefox", "CFBundleShortVersionString"),
    ("firefox-developer", "org.mozilla.firefoxdeveloperedition", "CFBundleShortVersionString"),
    ("firefox-nightly", "org.mozilla.nightly", "CFBundleShortVersionString"),
    ("safari", "com.apple.Safari", "CFBundleShortVersionString"),
    ("opera", "com.operasoftware.Opera", "CFBundleVersion"),
    ("opera-beta", "com.operasoftware.OperaNext", "CFBundleVersion"),
    ("opera-developer", "com.operasoftware.OperaDeveloper", "CFBundleVersion"),
    ("msedge", "com.microsoft.edgemac", "CFBundleVersion"),
    ("msedge-beta", "com.microsoft.edgemac.Beta", "CFBundleVersion"),
    ("msedge-dev", "com.microsoft.edgemac.Dev", "CFBundleVersion"),
    ("msedge-canary", "com.microsoft.edgemac.Canary", "CFBundleVersion"),
    ("brave", "com.brave.Browser", "CFBundleVersion"),
    ("brave-beta", "com.brave.Browser.beta", "CFBundleVersion"),
    ("brave-dev", "com.brave.Browser.dev", "CFBundleVersion"),
    ("brave-nightly", "com.brave.Browser.nightly", "CFBundleVersion"),
)


# get all installed browsers
def browsers() -> Iterator[Browser]:
    for browser, bundle_id, version_string in POSSIBLE_BROWSERS:
        paths = subprocess.getoutput(f'mdfind "kMDItemCFBundleIdentifier == {bundle_id}"').splitlines()
        for path in paths:
            with open(os.path.join(path, "Contents/Info.plist"), "rb") as f:
                plist = plistlib.load(f)
                executable_name = plist.get("CFBundleExecutable")
                executable = os.path.join(path, "Contents/MacOS", executable_name)
                description = plist.get("CFBundleDisplayName") or plist.get("CFBundleName", browser)
                version = plist[version_string]
                yield Browser(
                    name=browser,
                    description=description,
                    version=version,
                    location=executable if browser != "safari" else path
                )


# get details of a browser
def get_details_of(name) -> Optional[Browser]:
    for browser, bundle_id, version_string in (browser_record for browser_record in POSSIBLE_BROWSERS
                                               if browser_record[0] == name):
        paths = subprocess.getoutput(f'mdfind "kMDItemCFBundleIdentifier == {bundle_id}"').splitlines()
        for path in paths:
            with open(os.path.join(path, "Contents/Info.plist"), "rb") as f:
                plist = plistlib.load(f)
                executable_name = plist.get("CFBundleExecutable")
                executable = os.path.join(path, "Contents/MacOS", executable_name)
                description = plist.get("CFBundleDisplayName") or plist.get("CFBundleName", browser)
                version = plist[version_string]
                yield Browser(
                    name=browser,
                    description=description,
                    version=version,
                    location=executable if browser != "safari" else path
                )


# retrieve browser version
def get_version_of(name) -> Optional[Version]:
    for browser, bundle_id, version_string in (browser_record for browser_record in POSSIBLE_BROWSERS
                                               if browser_record[0] == name):
        paths = subprocess.getoutput(f'mdfind "kMDItemCFBundleIdentifier == {bundle_id}"').splitlines()
        for path in paths:
            with open(os.path.join(path, "Contents/Info.plist"), "rb") as f:
                plist = plistlib.load(f)
                version = plist[version_string]
                yield Version(
                    version=version
                )
    yield "Browser is not installed."


# check if the given browser is installed
def do_i_have_installed(name):
    for browser, bundle_id, version_string in (browser_record for browser_record in POSSIBLE_BROWSERS
                                               if browser_record[0] == name):
        paths = subprocess.getoutput(f'mdfind "kMDItemCFBundleIdentifier == {bundle_id}"').splitlines()
        for path in paths:
            with open(os.path.join(path, "Contents/Info.plist"), "rb") as f:
                plist = plistlib.load(f)
                executable_name = plist.get("CFBundleExecutable")
                if executable_name:
                    return True
    return False


# get default browser
def what_is_the_default_browser() -> Optional[str]:
    # cmd = "xdg-settings get default-web-browser".split()
    # if cmd:
    #     default_browser = subprocess.check_output(cmd, stderr=subprocess.DEVNULL).decode().strip()
    # else:
    #     default_browser = "No browser is set to default."
    # return default_browser
    return "This function is not yet supported."



import fnmatch
import sys
from typing import Iterator, Optional

from . import linux, mac, windows
from .common import Browser, Version, OS

__all__ = ["Browser",
           "browsers",
           "do_i_have_installed",
           "give_me_details_of",
           "get_version_of",
           "what_is_the_default_browser"]


def browsers() -> Iterator[Browser]:
    """
    Iterates over installed browsers.

    :return: Iterator of Tuple of browser key and browser information.
    """
    match sys.platform:
        case OS.LINUX:
            yield from linux.browsers()
        case OS.WINDOWS:
            yield from windows.browsers()
        case OS.MAC:
            yield from mac.browsers()
        case _:
            print("This operation system is not yet supported.")


def do_i_have_installed(name: str):
    """
    Checks if the provided browser is installed in system.

    Parameters:
                Possible browser entries / parameters\n
                for Linux:\n
                chrome, chromium, firefox, ms-edge, opera, opera-beta, opera-developer,
                brave, brave-beta, brave-nightly\n
                for macOS:\n
                chrome, chrome-canary, chromium, firefox, firefox-developer, firefox-nightly,
                safari, opera, opera-beta, opera-developer, ms-edge, ms-edge-beta, ms-edge-dev, ms-edge-canary,
                brave, brave-beta, brave-dev, brave-nightly\n
                for Windows:\n
                chrome, chrome-canary, chromium, firefox, firefox-developer, firefox-nightly,
                opera, opera-beta, opera-developer, ms-edge, ms-edge-beta, ms-edge-dev, ms-edge-canary, ms-ie,
                brave, brave-beta, brave-nightly
    :return: True or False depending on the browser is installed or not.
    """
    match sys.platform:
        case OS.LINUX:
            return linux.do_i_have_installed(name)
        case OS.MAC:
            return mac.do_i_have_installed(name)
        # case OS.WINDOWS: Windows is not yet supported.


def give_me_details_of(name: str) -> Optional[Browser]:
    """
    Retrieve browser details if the provided browser is installed in system.

    Parameters:
                Possible browser entries / parameters\n
                for Linux:\n
                chrome, chromium, firefox, ms-edge, opera, opera-beta, opera-developer,
                brave, brave-beta, brave-nightly\n
                for macOS:\n
                chrome, chrome-canary, chromium, firefox, firefox-developer, firefox-nightly,
                safari, opera, opera-beta, opera-developer, ms-edge, ms-edge-beta, ms-edge-dev, ms-edge-canary,
                brave, brave-beta, brave-dev, brave-nightly\n
                for Windows:\n
                chrome, chrome-canary, chromium, firefox, firefox-developer, firefox-nightly,
                opera, opera-beta, opera-developer, ms-edge, ms-edge-beta, ms-edge-dev, ms-edge-canary, ms-ie,
                brave, brave-beta, brave-nightly
    :return: Dictionary containing browser name, description, desktop version and location.
    """
    match sys.platform:
        case OS.LINUX:
            for found in (entity for entity in linux.get_details_of(name) if entity is not None):
                return found
        case OS.MAC:
            for found in (entity for entity in mac.get_details_of(name) if entity is not None):
                return found
        # case OS.WINDOWS: Windows is not yet supported.


def get_version_of(name: str) -> Optional[Version]:
    """
    Retrieve browser version if the provided browser is installed in system.

    Parameters:
                Possible browser entries / parameters\n
                for Linux:\n
                chrome, chromium, firefox, ms-edge, opera, opera-beta, opera-developer,
                brave, brave-beta, brave-nightly\n
                for macOS:\n
                chrome, chrome-canary, chromium, firefox, firefox-developer, firefox-nightly,
                safari, opera, opera-beta, opera-developer, ms-edge, ms-edge-beta, ms-edge-dev, ms-edge-canary,
                brave, brave-beta, brave-dev, brave-nightly\n
                for Windows:\n
                chrome, chrome-canary, chromium, firefox, firefox-developer, firefox-nightly,
                opera, opera-beta, opera-developer, ms-edge, ms-edge-beta, ms-edge-dev, ms-edge-canary, ms-ie,
                brave, brave-beta, brave-nightly
    :return: Browser description and version.
    """
    match sys.platform:
        case OS.LINUX:
            for found in (entity for entity in linux.get_version_of(name) if entity is not None):
                return found
        case OS.MAC:
            for found in (entity for entity in mac.get_version_of(name) if entity is not None):
                return found
        # case OS.WINDOWS: Windows is not yet supported.


def what_is_the_default_browser():
    """
    Shows the default browser in system - if there is any.

    :return: Default browser information.
    """
    match sys.platform:
        case OS.LINUX:
            return linux.what_is_the_default_browser()
        # case OS.MAC: Mac is not yet supported.
        #     return mac.what_is_the_default_browser()
        # case OS.WINDOWS: Windows is not yet supported.
